#!/usr/bin/env ruby
#
# Copyright 2008 Matthew Lee Hinman
#
# A POC fuzzy stream matcher, does fuzzy sorting.
# 
# Usage: ./fuzzymatch-sort.rb <pcapfile>

require './pcapparser.rb'
require './streambuilder.rb'

## The variation number to use
v = 15

p = PcapParser.new(File.new(ARGV[0]))

pcount = 0

## Our array of streams
$slist = []

## Fuzzy match, based on sequence number distance
def fuzzymatch(pkt, v)
  nnum = 0
  snum = 0
  sn = pkt.seq_num.to_i
  an = pkt.ack_num.to_i
  $slist.each { |str|
    snum = snum + 1
    str.sa_list.each { |num|
      nnum = nnum + 1
      ## If the sequence number is close enough
      if ((sn.to_i - num.to_i).abs) <= v && (sn != 0)
        ## Add the packet to the stream
        puts "seq num: #{sn} close enough to #{num} to add. Had to check #{snum} streams and #{nnum} seq/ack nums"
        str.add_packet(pkt)
        return
      ## Maybe the ack number is close enough?
      elsif ((an.to_i - num.to_i).abs <= v) && (an != 0)
        puts "ack num: #{an} close enough to #{num} to add. Had to check #{snum} streams and #{nnum} seq/ack nums"
        str.add_packet(pkt)
        return
      end
    }
  }
  puts "Starting a new stream..."
  s = Stream.new(pkt.ip_src,pkt.ip_dst)
  s.add_packet(pkt)
  # Add to the beginning of the list, so we don't have to list through all of them
  # This improved performance for large stream lists
  $slist.unshift(s)
end

def fuzzy_sort(s)
  r = {}
  s.each { |pkt|
    key = pkt.ack_num.to_s + pkt.seq_num.to_s
    key = key.to_i
    #puts "key- #{key}"
    
    # Check if the hash already has a packet for this size, and make sure that
    # we don't override a 'large' packet with a simple packet, this is sort of
    # a hack.
    if (!r[key].nil? && r[key].length > pkt.length)
      puts "key already exists and is longer than current packet"
      next
    end

    # This kills duplicate packets, because of the way that hashes work
    r[key] = pkt
  }

  # Sort by key, which will put it in correct order
  nr = []
  nr = r.sort
  packets = []

  nr.each { |k,v|
    #puts "key: #{k}, val: #{v.seq_num},#{v.ack_num}"
    packets.push(v)
  }

  return packets

end

def stream_sort(s)
  if (true) ## Debug output
    ## Print packets
    pcount = 0
    s.plist.each { |pkt|
      pcount = pcount + 1
      next unless pkt.tcp?

      print "[#{pcount}]\t"
      print_packet(pkt)
    }
  end

  print "\n"
  puts "==== Unsorted Streams ===="
  print "\n"
  
  ## Print out each hosts packets
  puts "Source\n-------"
  s.get_src_plist.each { |pkt|
    print_packet(pkt)
  }
  puts "Dest\n-------"
  s.get_dst_plist.each { |pkt|
    print_packet(pkt)
  }

  # Sort both the source and destination packet lists
  src_sorted_stream = fuzzy_sort(s.get_src_plist)
  dst_sorted_stream = fuzzy_sort(s.get_dst_plist)

  print "\n"
  puts "==== Sorted Streams ===="
  print "\n"

  puts "Source\n-------"
  src_sorted_stream.each { |pkt|
    print_packet(pkt)
  }

  puts "Dest\n-------"
  dst_sorted_stream.each { |pkt|
    print_packet(pkt)
  }
  
end

## Simply print the packet to stdout
def print_packet(pkt)
  return if pkt.nil?
  return unless pkt.tcp?
  print "[#{pkt.readable_flags}] "
  print [pkt.ip_src].pack("N").unpack("C4").join(".")
  print " -> "
  print [pkt.ip_dst].pack("N").unpack("C4").join(".")
  print "\tseq=#{pkt.seq_num} ack=#{pkt.ack_num} len=#{pkt.length}"
  print "\n"
end

## Start our program, printing packets and running them through fuzzymatch
p.each { |pkt|
  pcount = pcount + 1
  next unless pkt.tcp?

  print "[#{pcount}]\t"
  print_packet(pkt)

  fuzzymatch(pkt, v)
}

puts "Ended up with #{$slist.length} stream(s)."
i = 0
$slist.each { |s|
  i = i + 1
  puts "\n==> Stream #{i} contains #{s.pkt_num} packet(s)"
  puts "--> Sorting stream #{i}..."
  stream_sort(s)
}

